# Concurrency, Part 1
C++ program that demonstrates the use of std::async for parallel matrix multiplication, std::scoped_lock for data protection, and std::chrono for timing different numerical algorithms. Here's the implementation:


### Resources [(C++ Tutorial)](https://www.geeksforgeeks.org/c-plus-plus/?ref=shm)
[algorithm](../html_includes/algorithm.html), 
[auto](../html_src/auto.html), 
[command_line_arguments](../html_src/command_line_arguments.html),
[exception](../html_includes/exception.html), 
[iostream](../html_includes/iostream.html), 
[namespace](../html_src/namespace.html),
[range-based](../html_src/range-based.html),
[sstream](../html_includes/sstream.html), 
[std::cin](../html_includes/iostream.html), 
[std::cout](../html_includes/iostream.html), 
[std::cerr](../html_includes/iostream.html), 
[std::exception](../html_includes/exception.html), 
[std::ifstream](../html_includes/ifstream.html), 
[std::ofstream](../html_includes/ofstream.html), 
[stdexcept](../html_includes/stdexcept.html), 
[throw](../html_includes/exception.html),
[vector](../html_includes/vector.html)


### Implementation

```cpp
#include <iostream>
#include <vector>
#include <future>
#include <mutex>
#include <shared_mutex>
#include <chrono>
#include <random>
#include <algorithm>

using Matrix = std::vector<std::vector<double>>;

std::shared_mutex mtx;

// Function to generate a random matrix
Matrix generateRandomMatrix(int rows, int cols) {
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(0.0, 1.0);

    Matrix matrix(rows, std::vector<double>(cols));
    for (auto& row : matrix) {
        for (auto& elem : row) {
            elem = dis(gen);
        }
    }
    return matrix;
}

// Function to multiply two matrices
Matrix multiplyMatrices(const Matrix& A, const Matrix& B) {
    int m = A.size();
    int n = B[0].size();
    int p = B.size();

    Matrix result(m, std::vector<double>(n, 0.0));

    for (int i = 0; i < m; ++i) {
        for (int j = 0; j < n; ++j) {
            for (int k = 0; k < p; ++k) {
                result[i][j] += A[i][k] * B[k][j];
            }
        }
    }

    return result;
}

// Function to multiply a portion of matrices (for parallel execution)
void multiplyMatricesPortion(const Matrix& A, const Matrix& B, Matrix& result, int start_row, int end_row) {
    int n = B[0].size();
    int p = B.size();

    for (int i = start_row; i < end_row; ++i) {
        for (int j = 0; j < n; ++j) {
            double sum = 0.0;
            for (int k = 0; k < p; ++k) {
                sum += A[i][k] * B[k][j];
            }
            std::scoped_lock lock(mtx);
            result[i][j] = sum;
        }
    }
}

// Function to perform parallel matrix multiplication
Matrix parallelMultiplyMatrices(const Matrix& A, const Matrix& B) {
    int m = A.size();
    int n = B[0].size();

    Matrix result(m, std::vector<double>(n, 0.0));

    int num_threads = std::thread::hardware_concurrency();
    int rows_per_thread = m / num_threads;

    std::vector<std::future<void>> futures;

    for (int i = 0; i < num_threads; ++i) {
        int start_row = i * rows_per_thread;
        int end_row = (i == num_threads - 1) ? m : (i + 1) * rows_per_thread;

        futures.push_back(std::async(std::launch::async, multiplyMatricesPortion, 
                                     std::ref(A), std::ref(B), std::ref(result), start_row, end_row));
    }

    for (auto& future : futures) {
        future.wait();
    }

    return result;
}

// Function to measure execution time
template<typename Func, typename... Args>
auto measureTime(Func func, Args&&... args) {
    auto start = std::chrono::high_resolution_clock::now();
    auto result = func(std::forward<Args>(args)...);
    auto end = std::chrono::high_resolution_clock::now();

    std::chrono::duration<double, std::milli> duration = end - start;
    return std::make_pair(result, duration.count());
}

int main() {
    const int size = 1000;

    Matrix A = generateRandomMatrix(size, size);
    Matrix B = generateRandomMatrix(size, size);

    std::cout << "Matrix size: " << size << "x" << size << std::endl;

    auto [result_serial, time_serial] = measureTime(multiplyMatrices, A, B);
    std::cout << "Serial multiplication time: " << time_serial << " ms" << std::endl;

    auto [result_parallel, time_parallel] = measureTime(parallelMultiplyMatrices, A, B);
    std::cout << "Parallel multiplication time: " << time_parallel << " ms" << std::endl;

    // Check if results are the same
    bool results_match = (result_serial == result_parallel);
    std::cout << "Results match: " << (results_match ? "Yes" : "No") << std::endl;

    return 0;
}
```

This program demonstrates the following concepts:

1. **std::async and std::scoped_lock**:
   - We use `std::async` to launch parallel tasks for matrix multiplication in `parallelMultiplyMatrices`.
   - `std::scoped_lock` is used in `multiplyMatricesPortion` to protect shared data (the result matrix) during computation.

2. **std::shared_mutex**:
   - We use `std::shared_mutex` as a more flexible alternative to a regular mutex, allowing multiple readers or a single writer.

3. **std::chrono for timing**:
   - The `measureTime` function uses `std::chrono::high_resolution_clock` to measure the execution time of different algorithms (serial vs. parallel matrix multiplication in this case).

4. **Matrix Operations**:
   - We implement both serial and parallel versions of matrix multiplication to compare their performance.

5. **Random Matrix Generation**:
   - The `generateRandomMatrix` function creates random matrices for testing.

This program provides a practical example of using modern C++ features for parallel computation and performance measurement in a scientific computing context. It compares the execution time of serial and parallel matrix multiplication, demonstrating the potential performance benefits of parallelization for large-scale computations.

Remember that the actual performance gain from parallelization can vary depending on the hardware, matrix size, and other factors. For very small matrices or on systems with limited parallelism, the overhead of creating and managing threads might outweigh the benefits.

## Parallel code slower than serial code!
The default output of the example code suggest that the parallel code is *slower* than the serial code. 
Thank you for providing the output. You're correct in noting that the parallel version is slower than the serial version in this case. This is an interesting and not uncommon scenario in parallel computing. Let's analyze why this might be happening:

1. **Overhead of Thread Management**: Creating, managing, and synchronizing threads introduces overhead. For smaller matrices or less computationally intensive tasks, this overhead can outweigh the benefits of parallelization.  The results mentioned was with 12 threads (the number on my mac). When I reduced this to 2 threads, I got better than linear speedup. 

2. **Memory Access Patterns**: Matrix multiplication involves accessing large amounts of data. In a parallel version, multiple threads may be competing for memory access, potentially causing cache misses and memory contention.

3. **Synchronization Costs**: The use of `std::scoped_lock` in the parallel version introduces synchronization overhead. Each time a thread writes to the result matrix, it needs to acquire a lock, which can slow down the overall process.

4. **Load Balancing**: If the work is not evenly distributed among threads, some threads may finish early while others continue to work, reducing efficiency.

5. **Hardware Limitations**: The performance gain from parallelization is limited by the number of available CPU cores and the memory bandwidth of the system.

6. **Problem Size**: For smaller matrices, the benefits of parallelization may not be apparent. As the matrix size increases, you might see the parallel version outperform the serial version.

To potentially improve the parallel performance:

1. Increase the matrix size to make the computation more intensive relative to the threading overhead.
2. Use a more granular locking strategy or lock-free data structures to reduce synchronization overhead.
3. Implement cache-friendly algorithms to improve memory access patterns.
4. Experiment with different work distribution strategies to achieve better load balancing.
5. Consider using more specialized libraries for matrix operations that are optimized for parallel execution on specific hardware.

Remember, parallelization doesn't always lead to performance improvements, especially for smaller problems or when not properly optimized for the specific task and hardware. It's always important to measure and compare performance in real-world scenarios.

## Potential reasons given by Sonnet 3.5 for superlinear speedup
Thank you for providing that additional information. The better-than-linear speedup you observed when reducing the number of threads from 12 to 2 is indeed surprising and interesting. Let's analyze why this might have happened:

1. **Cache Efficiency**: With fewer threads, there's less competition for cache resources. Each thread can potentially use a larger portion of the cache, leading to fewer cache misses and better overall performance.

2. **Memory Bandwidth**: Matrix multiplication is often memory-bound rather than compute-bound. With 12 threads, you might have been saturating the memory bandwidth, causing contention. Two threads might be better matched to your system's memory bandwidth capabilities.

3. **NUMA Effects**: If your system has a Non-Uniform Memory Access (NUMA) architecture, using fewer threads might have resulted in better data locality, reducing the need for expensive cross-NUMA node memory accesses.

4. **Reduced Synchronization Overhead**: Fewer threads mean less synchronization is needed, which can significantly reduce overhead, especially if the locking mechanism was a bottleneck.

5. **Load Balancing**: With just two threads, it's easier to achieve good load balancing. With 12 threads, some might finish early and sit idle while others continue working.

6. **Hyperthreading**: If your CPU has 6 physical cores with hyperthreading (giving 12 logical cores), using 2 threads might be utilizing full physical cores more efficiently than when all 12 logical cores are in use.

7. **Reduced Context Switching**: Fewer threads can lead to less context switching overhead by the operating system.

8. **Better Alignment with Hardware Resources**: Two threads might align better with your specific hardware architecture, such as the number of memory channels or the layout of your CPU's cache hierarchy.

To further optimize and understand the performance characteristics:

1. Try varying the number of threads from 1 to your maximum logical core count, plotting the performance at each step.
2. Profile your code to identify bottlenecks, especially around memory access and synchronization points.
3. Experiment with different work distribution strategies and granularity of parallelism.
4. Consider using tools like Intel VTune or AMD μProf to get detailed hardware performance counters.

Achieving optimal performance in parallel computing often requires fine-tuning for the specific hardware and problem size. The "ideal" number of threads can vary based on these factors.

## Citations:
- [1] https://www.wolframalpha.com/input?input=performance+comparison+between+serial+and+parallel+matrix+multiplication
- [2] https://www.wolframalpha.com/input?input=advantages+of+parallel+processing+for+matrix+multiplication
- [3] https://www.wolframalpha.com/input?input=why+would+parallel+processing+be+slower+than+serial+processing%3F
- [4] https://www.wolframalpha.com/input?input=factors+affecting+performance+of+parallel+computing
- [5] https://www.wolframalpha.com/input?input=why+can+parallel+algorithms+be+slower+than+serial+algorithms
- [6] https://www.wolframalpha.com/input?input=overhead+in+parallel+processing
- [7] https://www.wolframalpha.com/input?input=thread+management+costs
- [8] https://www.wolframalpha.com/input?input=load+balancing+in+parallel+computing


## Related
- How can I optimize matrix multiplication using parallel processing with std::async
- What are the benefits of using std::scoped_lock over std::lock in multithreaded applications
- How do I measure the execution time of parallel tasks using std::chrono
- Can you provide an example of using std::shared_mutex for shared data protection
- How do I compare the efficiency of direct and iterative solvers using timers
